#include "flashimg.h"
#include "flashdisk.h"
#include "common.h"
#include "flashimgtool.h"

#define SECTOR_SIZE 512

static int flash_read_sector_oob(struct flash_disk *flash_disk,
                                 int sector, unsigned char *oob) {
    return flashimg_read(flash_disk->flash_img, oob,
                         sector / 4, 2048 + 16 * (sector % 4), 16);
}

static int flash_read_sector(struct flash_disk *flash_disk,
                             int sector,
                             unsigned char *buf, unsigned char *oob) {
    if (flashimg_read(flash_disk->flash_img, buf,
                      sector / 4, 512 * (sector % 4), 512) < 0)
        return -1;
    if (flashimg_read(flash_disk->flash_img, oob,
                      sector / 4, 2048 + 16 * (sector % 4), 16) < 0)
        return -1;
    return 0;
}

static int flash_program_page(struct flash_disk *flash_disk,
                              int sector,
                              const unsigned char *buf,
                              const unsigned char *oob) {
    if (flashimg_write(flash_disk->flash_img, buf,
                       sector / 4, 0, 2048) < 0)
        return -1;
    if (flashimg_write(flash_disk->flash_img, oob,
                       sector / 4, 2048, 64) < 0)
        return -1;
    return 0;
}

static int flash_update_program_page(struct flash_disk *flash_disk,
                                     int dest_sector,
                                     int upd_start, int upd_length,
                                     unsigned char *page,
                                     const unsigned char *buf,
                                     const unsigned char *oob) {
    if (upd_length) {
        memcpy(page + 512 * upd_start, buf, 512 * upd_length);
        memcpy(page + 2048 + 16 * upd_start, oob + 16 * upd_start,
               16 * upd_length);
    }
    return flash_program_page(flash_disk, dest_sector, page, page + 2048);
}

static int flash_copy_update_page(struct flash_disk *flash_disk,
                                  int src_sector, int dest_sector,
                                  int upd_start, int upd_length,
                                  const unsigned char *buf,
                                  const unsigned char *oob) {
    unsigned char page[2048 + 64];
    if (flashimg_read(flash_disk->flash_img, page,
                      src_sector / 4, 0, 2048 + 64) < 0)
        return -1;
    return flash_update_program_page(flash_disk, dest_sector,
                                     upd_start, upd_length,
                                     page, buf, oob);
}

static int flash_update_empty_page(struct flash_disk *flash_disk,
                                   int dest_sector,
                                   int upd_start, int upd_length,
                                   const unsigned char *buf,
                                   const unsigned char *oob) {
    unsigned char page[2048 + 64];
    memset(page, 0xff, sizeof(page));
    return flash_update_program_page(flash_disk, dest_sector,
                                     upd_start, upd_length,
                                     page, buf, oob);
}

static int flash_block_erase(struct flash_disk *flash_disk,
                             int sector) {
    int ps;
    unsigned char buf[2048 + 64];

    memset(buf, 0xff, sizeof(buf));
    for (ps = sector; ps < sector + 64 * 4; ps += 4)
        if (flash_program_page(flash_disk, ps, buf, buf + 2048) < 0)
            return -1;
    return 0;
}

static int flash_disk_get_n_segments(struct flash_disk *flash_disk)
{
    switch (flash_disk->model) {
        case FLASHIMG_256_8:
            return 2;
        case FLASHIMG_512_8:
            return 4;
        case FLASHIMG_1024_8:
            return 8;
    }
}

static inline int flash_disk_get_n_phblocks(struct flash_disk *flash_disk) {
    return 1024;
}

static inline int flash_disk_get_n_sectors_in_block(struct flash_disk
                                                    *flash_disk) {
    return 256;
}

static int flash_disk_phblock_to_sector(struct flash_disk *flash_disk,
                                        int segment, int block)
{
    return (segment * flash_disk_get_n_phblocks(flash_disk) + block)
        * flash_disk_get_n_sectors_in_block(flash_disk);
}

static int flash_disk_is_bad_block(unsigned char* oob)
{
    /* TODO: should we check two pages? (see datasheet) */
    return oob[0] != 0xff;
}

static int count_1(int n) {
    int r = 0;
    while (n != 0) {
        r += (n & 1);
        n >>= 1;
    }
    return r;
}

static int flash_disk_get_logical_block_no(unsigned char* oob)
{
    int no1, no2;
    no1 = oob[6] + (oob[7] << 8);
    no2 = oob[11] + (oob[12] << 8);

    if (no1 == no2 && (no1 & 0xf000) == 0x1000)
        return (no1 & 0xfff) >> 1;

    if (count_1(no1 ^ no2) > 1)
        return -1;

    if ((no1 & 0xf000) == 0x1000
        && (count_1(no1) & 1) == 0)
        return (no1 & 0xfff) >> 1;

    if ((no2 & 0xf000) == 0x1000
        && (count_1(no2) & 1) == 0)
        return (no2 & 0xfff) >> 1;

    return -1;
}

static void flash_disk_set_oob(struct flash_disk *flash_disk,
                               char *oob, int sector)
{
    int blockmod, enc;
    oob[0] = oob[5] = 0xff;
    oob[1] = 0x5a;

    blockmod = (sector / flash_disk_get_n_sectors_in_block(flash_disk))
        % SEGMENT_SIZE;
    enc = 0x1000 + (blockmod << 1);
    if (count_1(enc) & 1) enc |= 1;
    oob[6] = oob[11] = enc & 0xff;
    oob[7] = oob[12] = enc >> 8;
}

int flash_disk_scan(struct flash_disk *flash_disk)
{
    int n_segments, n_phblocks;
    unsigned char oob[16];
    int s, b;
    
    /* TODO: checking for double blocks */

    n_segments = flash_disk_get_n_segments(flash_disk);
    n_phblocks = flash_disk_get_n_phblocks(flash_disk);

    flash_disk->cur_block = -1;
    flash_disk->cur_phblock_start = -1;

    for (s = 0; s < n_segments; s++)
    {
        for (b = 0; b < n_phblocks; b++)
        {
            int r;
            r = flash_read_sector_oob(flash_disk,
                                      flash_disk_phblock_to_sector(flash_disk,
                                                                   s, b),
                                      oob);
            if (r >= 0 && !flash_disk_is_bad_block(oob))
            {
                int lb;
                lb = flash_disk_get_logical_block_no(oob);
                if (lb >= 0 && lb < SEGMENT_SIZE) {
                    DEBUGF("logical %d - physical %d\n", lb, b);
                    flash_disk->block_map[s][lb] = b;
                }
            }
        }
    }
    return 0;
}

int flash_disk_find_block(struct flash_disk *flash_disk, int block)
{
    int seg, bmod, phb;
    unsigned char oob[16];
    int r;
   
    if (block >= SEGMENT_SIZE * flash_disk_get_n_segments(flash_disk))
        return -1;

    if (block == flash_disk->cur_block)
        return flash_disk->cur_phblock_start;

    seg = block / SEGMENT_SIZE;
    bmod = block % SEGMENT_SIZE;

    phb = flash_disk->block_map[seg][bmod];
    r = flash_read_sector_oob(flash_disk,
                              flash_disk_phblock_to_sector(flash_disk,
                                                           seg, phb),
                              oob);
    if (r < 0)
        return -1;
    if (flash_disk_is_bad_block(oob))
        return -1;
    if (flash_disk_get_logical_block_no(oob) != bmod)
        return -1;

    flash_disk->cur_block = block;
    flash_disk->cur_phblock_start = flash_disk_phblock_to_sector(flash_disk,
                                                                 seg, phb);
    return flash_disk->cur_phblock_start;
}

int flash_disk_find_free_block(struct flash_disk *flash_disk, int block)
{
    int seg, b;
    unsigned char oob[16];
   
    if (block >= SEGMENT_SIZE * flash_disk_get_n_segments(flash_disk))
        return -1;

    seg = block / SEGMENT_SIZE;
    /* TODO: wear leveling: start searching from random block?
     * continue where the last free block was found? */
    for (b = 0; b < flash_disk_get_n_phblocks(flash_disk); b++) {
        int r, s;
        s = flash_disk_phblock_to_sector(flash_disk, seg, b);
        r = flash_read_sector_oob(flash_disk, s, oob);
        if (!flash_disk_is_bad_block(oob) && oob[1] == 0xff)
            return s;
    }

    return -1;
}

int flash_disk_read_sectors(struct flash_disk *flash_disk,
                            unsigned long start,
                            int count,
                            void* buf)
{
    int block, secmod, done;
    int phb;
    char oob[16];

    block = start / flash_disk_get_n_sectors_in_block(flash_disk);
    secmod = start % flash_disk_get_n_sectors_in_block(flash_disk);

    phb = flash_disk_find_block(flash_disk, block);
    done = 0;
    while (count > 0 && secmod < flash_disk_get_n_sectors_in_block(flash_disk))
    {
        if (phb >= 0)
            flash_read_sector(flash_disk, phb + secmod, buf, oob);
        else
            memset(buf, 0, SECTOR_SIZE);

        buf += SECTOR_SIZE;
        count--;
        secmod++;
        done++;
    }
    return done;
}

int flash_disk_write_sectors(struct flash_disk *flash_disk,
                             unsigned long start,
                             int count,
                             const void* buf)
{
    int srcph, destph, block, secmod;
    int ps, i, r, done;
    char oob[64];

    block = start / flash_disk_get_n_sectors_in_block(flash_disk);
    secmod = start % flash_disk_get_n_sectors_in_block(flash_disk);

    srcph = flash_disk_find_block(flash_disk, block);
    destph = flash_disk_find_free_block(flash_disk, block);
    if (destph < 0)
        return -1;

    done = 0;
    for (ps = 0; ps < flash_disk_get_n_sectors_in_block(flash_disk); ps += 4)
    {
        if (ps >= secmod && ps + 3 < secmod + count)
        {
            for (i = 0; i < 4; i++)
                flash_disk_set_oob(flash_disk, oob + i * 16, start + done + i);
            r = flash_program_page(flash_disk, destph + ps, buf, oob);
            /* TODO: mark as bad and try again */
            if (r < 0)
                return -1;
            done += 4;
            buf += SECTOR_SIZE * 4;
        }
        else
        {
            int a, b, n;
            a = secmod - ps;
            if (a < 0) a = 0;
            b = secmod + count - ps;
            if (b > 4) b = 4;
            n = b - a;
            if (n < 0) n = 0;
            for (i = 0; i < 4; i++)
                flash_disk_set_oob(flash_disk,
                                   oob + i * 16, start + done + i - a);
            if (srcph >= 0)
                r = flash_copy_update_page(flash_disk,
                                           srcph + ps, destph + ps,
                                           a, n, buf, oob);
            else
                r = flash_update_empty_page(flash_disk,
                                            destph + ps, a, n, buf, oob);
            /* TODO: mark as bad and try again */
            if (r < 0)
                return -1;
            done += n;
            buf += SECTOR_SIZE * n;
        }
    }

    if (srcph >= 0)
        if (flash_block_erase(flash_disk, srcph) < 0) {
            /* TODO: mark as bad */
        }

    flash_disk->block_map[block / SEGMENT_SIZE][block % SEGMENT_SIZE] =
        (destph / flash_disk_get_n_sectors_in_block(flash_disk))
        % flash_disk_get_n_phblocks(flash_disk);
    flash_disk->cur_block = block;
    flash_disk->cur_phblock_start = destph;
    return done;
}

struct flash_disk *flash_disk_create(struct flashimg *flash_img) {
    struct flash_disk *flash_disk;
    flash_disk = xmalloc(sizeof(struct flash_disk));
    if (!flash_disk)
        return NULL;

    flash_disk->flash_img = flash_img;
    flash_disk->model = flash_img->model;
    
    if (flash_disk_scan(flash_disk) < 0) {
        flash_disk_free(flash_disk);
        return NULL;
    }
    return flash_disk;
}

void flash_disk_free(struct flash_disk *flash_disk) {
    flashimg_free(flash_disk->flash_img);
    free(flash_disk);
}
