#include <stdarg.h>
#include <stdio.h>
#include "common.h"
#include "disk.h"
#include "flashimg.h"
#include "flashdisk.h"
#include "ata.h"
#include "dir.h"

static struct flash_disk *flash_disk;
static int option_debug = 0;

void debugf(const char *format, ...) {
  if (option_debug) {
    va_list ap;
    fprintf(stderr, "debug: ");
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
  }
}

int init_flash(char *name) {
  struct flashimg *flash_img;

  flash_img = flashimg_create(FLASHIMG_256_8, name, 0);
  if (!flash_img)
    return -1;
  flash_disk = flash_disk_create(flash_img);
  if (!flash_disk) {
    flashimg_free(flash_img);
    return -1;
  }
  ata_init(flash_disk);
  return 0;
}

int cmd_ls(int argc, char *argv[]) {
  DIR* dir;
  struct dirent* entry;

  if (argc < 1)
    return 1;
  dir = opendir(argv[0]);
  if (dir) {
    while ((entry = readdir(dir)))
      printf("%-15s %10d\n", entry->d_name, entry->size);
    closedir(dir);
    return 0;
  } else {
    error("could not open dir %s", argv[0]);
    return 1;
  }
}

int cmd_mkdir(int argc, char *argv[]) {
  if (argc < 1)
    return 1;
  if (mkdir(argv[0], 0) < 0) {
    error("could not create dir %s", argv[0]);
    return 1;
  }
  return 0;
}

int cmd_get(int argc, char *argv[]) {
  int fd;
  FILE *fp;
  char buf[4096];

  if (argc < 2)
    return 1;
  fd = open(argv[0], O_RDONLY);
  if (fd < 0) {
    error("could not open file %s", argv[0]);
    return 1;
  }
  fp = fopen(argv[1], "w");
  if (!fp) {
    close(fd);
    error_sys("could not open file %s", argv[1]);
    return 1;
  }
  while (1) {
    int r = read(fd, buf, sizeof(buf));
    if (r < 0) {
      close(fd);
      fclose(fp);
      error("error while reading %s", argv[0]);
      return 1;
    }
    if (r == 0) break;
    r = fwrite(buf, r, 1, fp);
    if (r != 1) {
      close(fd);
      fclose(fp);
      error("error while writing %s", argv[1]);
      return 1;
    }
  }
  close(fd);
  fclose(fp);
  return 0;
}

int cmd_put(int argc, char *argv[]) {
  int fd;
  FILE *fp;
  char buf[4096];

  if (argc < 2)
    return 1;
  fd = creat(argv[1], O_WRONLY);
  if (fd < 0) {
    error("could not open file %s", argv[1]);
    return 1;
  }
  fp = fopen(argv[0], "r");
  if (!fp) {
    close(fd);
    error_sys("could not open file %s", argv[0]);
    return 1;
  }
  do {
    int rw;
    int r = fread(buf, 1, sizeof(buf), fp);
    if (ferror(fp)) {
      close(fd);
      fclose(fp);
      error("error while reading %s", argv[0]);
      return 1;
    }
    if (r) {
      rw = write(fd, buf, r);
      if (rw != r) {
	close(fd);
	fclose(fp);
	error("error while writing %s", argv[1]);
	return 1;
      }
    }
  } while (!feof(fp));
  close(fd);
  if (fclose(fp) == EOF) {
    error_sys("error while closing %s", argv[1]);
    return 1;
  }
  return 0;
}

struct {
  char *cmd;
  int (*func)(int argc, char* argv[]);
} func_tab[] = {
  {"ls", cmd_ls},
  {"mkdir", cmd_mkdir},
  {"get", cmd_get},
  {"put", cmd_put},
  /*    {"rm", cmd_rm},
   */    {NULL, NULL}
};

int main(int argc, char *argv[]) {
  int i;
  set_program_name(argv[0]);

  if (argc < 3)
    return 1;

  if (init_flash(argv[1]) < 0)
    return 1;
  if (disk_mount_all() <= 0) {
    printf("mount failed\n");
    return 1;
  }
  for (i = 0; func_tab[i].cmd; i++)
    if (!strcmp(argv[2], func_tab[i].cmd))
      return (*func_tab[i].func)(argc - 3, argv + 3);

  return 0;
}
