/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdlib.h>

#include "usb.h"
#include "log.h"

struct reg_structure {
  int bit;
  int n;
  char *name;
};

struct reg_structure mode_structure[] = {
  {0, 1, "SOFTCT"},
  {1, 1, "PWRON"},
  {2, 1, "WKUPCS"},
  {3, 1, "GLINTENA"},
  {4, 1, "SFRESET"},
  {5, 1, "GOSUSP"},
  {6, 1, "SNDRSU"},
  {7, 1, "CLKAON"},
  {9, 1, "DMACLKON"},
  {-1, -1, NULL}
};

struct reg_structure inten_structure[] = {
  {0, 1, "BRST"},
  {1, 1, "SOF"},
  {2, 1, "PSOF"},
  {3, 1, "SUSP"},
  {4, 1, "RESM"},
  {5, 1, "HS_STA"},
  {6, 1, "DMA"},
  {7, 1, "VBUS"},
  {8, 1, "EP0SETUP"},
  {10, 1, "EP0RX"},
  {11, 1, "EP0TX"},
  {12, 1, "EP1RX"},
  {13, 1, "EP1TX"},
  {14, 1, "EP2RX"},
  {15, 1, "EP2TX"},
  {16, 1, "EP3RX"},
  {17, 1, "EP3TX"},
  {18, 1, "EP4RX"},
  {19, 1, "EP4TX"},
  {20, 1, "EP5RX"},
  {21, 1, "EP5TX"},
  {22, 1, "EP6RX"},
  {23, 1, "EP6TX"},
  {24, 1, "EP7RX"},
  {25, 1, "EP7TX"},
  {-1, -1, NULL}
};

struct reg_structure intconf_structure[] = {
  {0, 1, "INTPOL"},
  {1, 1, "INTLVL"},
  {2, 2, "DDBGMODOUT"},
  {4, 2, "DDBGMODIN"},
  {6, 2, "CDBGMOD"},
  {-1, -1, NULL}
};

struct reg_structure epindex_structure[] = {
  {0, 1, "DIR"},
  {1, 4, "ENDPIDX"},
  {5, 1, "EP0SETUP"},
  {-1, -1, NULL}
};

struct reg_structure ctrl_structure[] = {
  {0, 1, "STALL"},
  {1, 1, "STATUS"},
  {2, 1, "DSEN"},
  {3, 1, "VENDP"},
  {4, 1, "CLBUF"},
  {-1, -1, NULL}
};

struct reg_structure maxpacketsize_structure[] = {
  {0, 11, "FFOSZ"},
  {11, 2, "NTRANS"},
  {-1, -1, NULL}
};

struct reg_structure type_structure[] = {
  {0, 2, "ENDTYP"},
  {2, 1, "DBLBUF"},
  {3, 1, "ENABLE"},
  {4, 1, "NOEMPKT"},
  {-1, -1, NULL}
};

static void bit_values(char *s, uint32 val, struct reg_structure *reg) {
  int first = 1;
  int n;
  while (reg->bit >= 0) {
    if (first)
      first = 0;
    else
      *s++ = ' ';
    sprintf(s, "%s=%x%n",
	    reg->name, (val >> reg->bit) & ((1 << reg->n) - 1), &n);
    s += n;
    reg++;
  }
  *s = 0;
}

static int fill_in_fifo(struct usb *usb, struct usb_in_fifo *in_fifo) {
  if (in_fifo->len > 0)
    return USB_ERR_BUFFER_FULL;

  in_fifo->ptr = 0;
  in_fifo->len = usb->in_buf_ptr;
  memcpy(in_fifo->data, usb->in_buf, usb->in_buf_ptr);
  usb->in_buf_ptr = 0;
  return 0;
}

static int in_fifo_get_byte(struct usb_in_fifo *in_fifo) {
  if (in_fifo->len) {
    int r = in_fifo->data[in_fifo->ptr++];
    if (in_fifo->ptr == in_fifo->len)
      in_fifo->len = 0;
    return r;
  } else
    return 0;
}

static int in_fifo_get_half_word(struct usb_in_fifo *in_fifo) {
  int l, h;
  l = in_fifo_get_byte(in_fifo);
  h = in_fifo_get_byte(in_fifo);
  return (h << 8) | l;
}

struct usb *create_usb(void) {
  struct usb *usb;

  usb = xmalloc(sizeof(struct usb));
  if (!usb)
    return NULL;

  usb_reset(usb);

  return usb;
}

void free_usb(struct usb *usb) {
  if (usb) free(usb);
}

void usb_reset(struct usb *usb) {
  int i;

  usb->mode_reg = 0;
  usb->inten_reg = 0;
  usb->intconf_reg = 0xfc;
  usb->index_reg = 0;

  for (i = 0; i < 8; i++) {
    usb->endpoint[i].ctrl = 0;
    usb->endpoint[i].buflen = 0;
    usb->endpoint[i].maxpacketsize = 0;
    usb->endpoint[i].type = 0;
    usb->endpoint[i].in_fifo.len = 0;
    usb->endpoint[i].out_ptr = 0;
  }

  usb->in_buf_ptr = 0;
}

static int usb_dir(struct usb *usb) {
  return usb->index_reg & 1;
}

static int usb_setup(struct usb *usb) {
  return (usb->index_reg >> 5) & 1;
}

static int endpoint_index(struct usb *usb) {
  int n = (usb->index_reg >> 1) & 1;
  return (n > 7) ? 7 : n;
}

static struct usb_endpoint *endpoint(struct usb *usb) {
  return usb->endpoint + endpoint_index(usb);
}

uint32 usb_read_word(void *data, uint32 adr) {
  struct usb *usb = data;

  switch (adr & 0xffff) {
  case 0x14:
    logmsg(LOG_USB, 10, "usb: read interrupt enable\n");
    return usb->inten_reg;
  case 0x18:
    logmsg(LOG_USB, 10, "usb: read interrupt\n");
    return usb->int_status;
  case 0x70:
    logmsg(LOG_USB, 10, "usb: read id\n");
    return 0x158230;
  default:
    logmsg(LOG_USB, 5, "usb: read word from %x\n", adr);
    return 0;
  }
}

void usb_write_word(void *data, uint32 adr, uint32 val) {
  char tmp[200];
  struct usb *usb = data;

  switch (adr & 0xffff) {
  case 0x14:
    usb->inten_reg = val;
    bit_values(tmp, val, inten_structure);
    logmsg(LOG_USB, 10, "usb: write %x to inten (%s)\n", val, tmp);
    break;
  case 0x18:
    logmsg(LOG_USB, 10, "usb: write %x to interrupt\n", val);
    usb->int_status &= ~val;
    break;
  default:
    logmsg(LOG_USB, 5, "usb: write word %x to %x\n", val, adr);
  }
}

uint32 usb_read_half_word(void *data, uint32 adr) {
  struct usb *usb = data;

  switch (adr & 0xffff) {
  case 0x04:
    logmsg(LOG_USB, 10, "usb: read from maxpacketsize\n");
    return endpoint(usb)->maxpacketsize;
  case 0x8:
    logmsg(LOG_USB, 10, "usb: read from endpoint type\n");
    return endpoint(usb)->type;
  case 0x0c:
    logmsg(LOG_USB, 10, "usb: read from mode reg\n");
    return usb->mode_reg;
  case 0x1c:
    logmsg(LOG_USB, 10, "usb: read from buflen\n");
    if (usb_setup(usb))
      return usb->setup_in_fifo.len;
    else if (usb_dir(usb) == 0)
      return endpoint(usb)->in_fifo.len;
    else
      return endpoint(usb)->buflen;
  case 0x20:
    logmsg(LOG_USB, 10, "usb: read from data port\n");
    if (usb_setup(usb))
      return in_fifo_get_half_word(&usb->setup_in_fifo);
    else if (usb_dir(usb) == 0)
      return in_fifo_get_half_word(&endpoint(usb)->in_fifo);
    else
      return 0;
  default:
    logmsg(LOG_USB, 5, "usb: read half word from %x\n", adr);
    return 0;
  }
}

static void usb_send_packet(struct usb *usb) {
  struct usb_endpoint *ep = endpoint(usb);
  char tmp[USB_BUF_SIZE * 3 + 1];
  int i;

  for (i = 0; i < ep->out_ptr; i++)
    sprintf(tmp + 3 * i, " %02x", ep->out_buf[i]);
  tmp[3 * ep->out_ptr] = 0;
  logmsg(LOG_USB, 5, "usb: sending packet:%s\n", tmp);

  usb->int_status |= (1 << (11 + endpoint_index(usb) * 2));
  ep->out_ptr = 0;
}

static void usb_write_to_fifo(struct usb *usb, uint32 val) {
  struct usb_endpoint *ep = endpoint(usb);
  ep->out_buf[ep->out_ptr++] = val & 0xff;
  if (ep->out_ptr >= ep->buflen)
    usb_send_packet(usb);
  else {
    ep->out_buf[ep->out_ptr++] = (val >> 8) & 0xff;
    if (ep->out_ptr >= ep->buflen)
      usb_send_packet(usb);
  }
}

void usb_write_half_word(void *data, uint32 adr, uint32 val) {
  struct usb *usb = data;
  char tmp[100];

  switch (adr & 0xffff) {
  case 0x04:
    endpoint(usb)->maxpacketsize = val;
    /* TODO reinitialize buflens */
    bit_values(tmp, val, maxpacketsize_structure);
    logmsg(LOG_USB, 10, "usb: maxpacketsize set to %x (%s)\n", val, tmp);
    break;
  case 0x08:
    endpoint(usb)->type = val;
    bit_values(tmp, val, type_structure);
    logmsg(LOG_USB, 10, "usb: write %x to endpoint type (%s)\n", val, tmp);
    break;
  case 0x0c:
    usb->mode_reg = val;
    bit_values(tmp, val, mode_structure);
    logmsg(LOG_USB, 10, "usb: mode reg set to %x (%s)\n", val, tmp);
    break;
  case 0x1c:
    endpoint(usb)->buflen = val;
    logmsg(LOG_USB, 10, "usb: write %x to buflen\n", val);
    break;
  case 0x20:
    logmsg(LOG_USB, 10, "usb: write %x to data port\n", val);
    usb_write_to_fifo(usb, val);
    break;
  case 0x7c:
    if (val == 0xaa37)
      logmsg(LOG_USB, 10, "usb: device unlocked\n");
    else
      logmsg(LOG_USB, 5, "usb: invalid vaule written to unlock\n");
    break;
  default:
    logmsg(LOG_USB, 5, "usb: write half word %x to %x\n", val, adr);
  }
}

uint8 usb_read(void *data, uint32 adr) {
  struct usb *usb = data;

  switch (adr & 0xffff) {
  case 0x10:
    logmsg(LOG_USB, 10, "usb: read from interrupt configuration\n");
    return usb->intconf_reg;
  case 0x1e:
    logmsg(LOG_USB, 10, "usb: read from buffer status\n");
    return 0;
  case 0x28:
    logmsg(LOG_USB, 10, "usb: read from control function\n");
    return endpoint(usb)->ctrl;
  default:
    logmsg(LOG_USB, 5, "usb: read byte from %x\n", adr);
    return 0;
  }
}

void usb_write(void *data, uint32 adr, uint8 val) {
  struct usb *usb = data;
  char tmp[100];

  switch (adr & 0xffff) {
  case 0x0:
    logmsg(LOG_USB, 10, "usb: write %x to address\n", val);
    break;
  case 0x10:
    usb->intconf_reg = val;
    bit_values(tmp, val, intconf_structure);
    logmsg(LOG_USB, 10, "usb: write %x to intconf (%s)\n", val, tmp);
    break;
  case 0x28:
    endpoint(usb)->ctrl = val & ~0x18;
    if (val & 0x10) {
      if (usb_setup(usb))
	usb->setup_in_fifo.len = 0;
      else
	endpoint(usb)->in_fifo.len = 0;
    }
    if (val & 0x80) {

    }
    bit_values(tmp, val, ctrl_structure);
    logmsg(LOG_USB, 10, "usb: write %x to control function (%s)\n", val, tmp);
    break;
  case 0x2c:
    usb->index_reg = val;
    bit_values(tmp, val, epindex_structure);
    logmsg(LOG_USB, 10, "usb: write %x to endpoint index (%s)\n", val, tmp);
    break;
  default:
    logmsg(LOG_USB, 5, "usb: write byte %x to %x\n", val, adr);
  }
}

void usb_receive_byte(struct usb *usb, uint8 byte) {
  if (usb->in_buf_ptr < USB_BUF_SIZE)
    usb->in_buf[usb->in_buf_ptr++] = byte;
}

int usb_bus_reset(struct usb *usb) {
  usb->int_status |= 1;
  return 0;
}

int usb_suspend(struct usb *usb) {
  usb->int_status |= 8;
  return 0;
}

int usb_resume(struct usb *usb) {
  usb->int_status |= 0x10;
  return 0;
}

int usb_hs_stat(struct usb *usb) {
  usb->int_status |= 0x20;
  return 0;
}

int usb_rcv_setup(struct usb *usb) {
  int r;
  r = fill_in_fifo(usb, &(usb->setup_in_fifo));
  if (r < 0) return r;
  usb->int_status |= 0x100;
  return 0;
}

int usb_rcv_data(struct usb *usb, int e) {
  int r;
  r = fill_in_fifo(usb, &(usb->endpoint[e].in_fifo));
  if (r < 0) return r;
  usb->int_status |= (1 << (10 + 2 * e));
  return 0;
}
